#
# mailest graphical user interface using Gtk+
#
#     Copyright(c) 1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

require 'est/estcore'
require 'gtk'



def printerr_full( err, io = $stderr )
  io.print err.to_s, " (#{err.type})\n"
  err.backtrace.each{|i| io.puts i }
end


def protect( place )
  begin
    ret = yield
  rescue
    $stderr.puts 'Error in ' + place if place
    printerr_full $!, $stderr
		return nil
  end

  ret
end



module EST

	Interface = 'mailest'
	Interface.freeze


  class Core

    private

    def initialize
      @working       = false
      @shutting_down = false

      @editers = []
      @addrbook = nil
    end

    def init_rc
      super
      @rc['if']['libdir'] = @rc['core']['libdir'] + '/' + 'mailest'
    end


    public


    attr :working

    def trigger
      return false if @working
      @working = true

      do_initialize
      @mailset.init_after

      notice "mailest ready"
      Gtk.main
    end


    def notice( str )
      mailset.write str
    end

    
    def address_book
      unless @addrbook then
        @addrbook = AddressBook.new( @rc['core']['addrfile'] )
      end
      @addrbook
    end

    def edit( port )
      unless @shutting_down then
        rc = CORE.rc['if']['editer']
        cls = (rc['type'] == 'internal') ? InternalEditer : ExternalEditer
        @editers.push cls.new( port, rc )
        return true
      else
        return false
      end
    end

    def edit_done( ed )
      unless @shutting_down then
        @editers.delete( ed )
      end
    end

    def editers
      @editers.dup
    end


    def shutdown( force = false )
      return if @shutting_down
      @shutting_down = true
      unless force then
        unless quit_ok? then
          @shutting_down = false
          return
        end
      end

      # shutdown realy
      notice 'going to shutdown...'
      @editers.each do |ed| ed.shutdown end
      @editers.clear
      super()
      exit 0
      @shutting_down = false    # not reach
    end


    private

    def quit_ok?
      true
    end

  end   # class Core



  ####
	#### mailset
	####

  class Mailset

    def initialize
      super

      construct_gui
      @inited = false
    end

    def init_after
      return if @inited
      @inited = true

      @current = self['inbox']
      @win.show_all
      @current.select 0
    end


    def shutdown
      super
      @win.destroy
    end


    def view( mail )
      upper = CORE.composer.view_upper( mail )
      lower = CORE.composer.view_lower( mail )

      @text.freeze
      @text.set_point 0
      @text.forward_delete @text.get_length

      @text.insert_text( upper, 0 )
      @text.thaw
      val = @textadj.upper

      @text.freeze
      @text.insert_text( "\n", @text.get_length )
      @text.insert_text( lower, @text.get_length )
      @text.thaw
      @textadj.value = val
    end


    def write( str )
      @notice.set str.to_s
    end


    def chmbox( name )
      name.strip!
      return if @current.name == name
      return unless tmp = get_targetmbox( name )

      # keep mbox single select mode
      if @current.multiselect then
        @excheck.clicked
      end

      old = @current
      @current = tmp
      @current.show
      old.hide
      
      @current.select @current.selected

      if @current.size == 0 then
        CORE.notice "no mail"
      else
        CORE.notice "#{@current.size} mails"
      end
    end


    def discard_mail( mbox, idx )
      if mbox.name == 'trash' then
        rmmail( 'trash', idx )
      else
        mvmail( mbox.name, idx, 'trash' )
      end
      if idx >= mbox.size then
        idx = mbox.size - 1
      end
      mbox.select idx
    end


    def read_ahead
      # 90% ahead
      val = @textadj.value + @textadj.page_size * 9 / 10

      if val >= @textadj.upper then   # at tail
        @current.view_next
      else
        @textadj.value = [ val, @textadj.upper - 1 ].min
      end
    end


    def read_back
      if @textadj.value <= @textadj.lower then   # at head
        @current.view_prev
      else
        # 90% back
        @textadj.value = [ @textadj.value - @textadj.page_size * 9 / 10,
                           @textadj.lower ].max
      end
    end


    private


    def construct_gui
      @win = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest' )
      @win.border_width(0)
      @win.set_usize( 640, 480 )

      @bigbox   = Gtk::VBox.new( false, 0 )
      @win.add @bigbox

      @topbox   = Gtk::HBox.new( false, 0 )
      @bigbox.pack_start( @topbox, false, false, 0 )

      @paned = Gtk::VPaned.new
      @paned.handle_size 7
      @paned.gutter_size 7
      @bigbox.pack_start @paned, true, true, 0

      @excheck = Gtk::CheckButton.new('')
      @topbox.pack_start @excheck, false, false, 0

      @mboxcombo = Gtk::Combo.new
      @mboxcombo.set_usize 100, 24
      @mboxcombo.entry.set_editable false
      @mboxcombo.set_popdown_strings mbox_names()
      @mboxcombo.entry.set_text 'inbox'
      @topbox.pack_start @mboxcombo, false, false, 0

      mrc = CORE.rc['if']['menu']
      @menubar             = Gtk::MenuBar.new
      @topbox.pack_start( @menubar, true, true, 0 )
        @item_move         = Gtk::MenuItem.new( mrc['move'] )
        @menumove          = Gtk::Menu.new
        @menubar.append @item_move
        @item_move.set_submenu @menumove

        @item_trash        = Gtk::MenuItem.new( mrc['trash'] )
        @menubar.append @item_trash
 
=begin c
        @item_mbox         = Gtk::MenuItem.new( mrc['mbox'] )
        @menubar.append @item_mbox
        @menumbox          = Gtk::Menu.new
        @item_mbox.set_submenu @menumbox
          @item_clear      = Gtk::MenuItem.new( mrc['clear'] )
          @menumbox.add @item_clear
          @item_update     = Gtk::MenuItem.new( mrc['update'] )
          @menumbox.add @item_update
=end c
          @item_ahead      = Gtk::MenuItem.new( '' )
          @item_back       = Gtk::MenuItem.new( '' )
          @item_next       = Gtk::MenuItem.new( '' )
          @item_prev       = Gtk::MenuItem.new( '' )
					@item_prevpage   = Gtk::MenuItem.new( '' )
					@item_nextpage   = Gtk::MenuItem.new( '' )

        @item_mail         = Gtk::MenuItem.new( mrc['mail'] )
        @menubar.append @item_mail
        @menumail          = Gtk::Menu.new
        @item_mail.set_submenu @menumail
          @item_reply      = Gtk::MenuItem.new( mrc['reply'] )
          @menumail.add @item_reply
          @item_write      = Gtk::MenuItem.new( mrc['write'] )
          @menumail.add @item_write
          @item_reedit     = Gtk::MenuItem.new( mrc['reedit'] )
          @menumail.add @item_reedit
          @item_fwd        = Gtk::MenuItem.new( mrc['forward'] )
          @menumail.add @item_fwd
          @item_clear      = Gtk::MenuItem.new( mrc['clear'] )
          @menumail.add @item_clear
          @item_update     = Gtk::MenuItem.new( mrc['update'] )
          @menumail.add @item_update

=begin c
        @item_addr         = Gtk::MenuItem.new( mrc['addr'] )
        @menubar.append @item_addr
        @menuaddr          = Gtk::Menu.new
        @item_addr.set_submenu @menuaddr
          @item_editaddr   = Gtk::MenuItem.new( mrc['editaddr'] )
          @menuaddr.add @item_editaddr
=end c

        @item_inet         = Gtk::MenuItem.new( mrc['inet'] )
        @menubar.append @item_inet
        @menuinet          = Gtk::Menu.new
        @item_inet.set_submenu @menuinet
          @item_pop        = Gtk::MenuItem.new( mrc['pop'] )
          @menuinet.add @item_pop
          @item_send       = Gtk::MenuItem.new( mrc['send'] )
          @menuinet.add @item_send

        @item_app          = Gtk::MenuItem.new( mrc['app'] )
        @menubar.append @item_app
        @menuapp           = Gtk::Menu.new
        @item_app.set_submenu @menuapp
          @item_editaddr   = Gtk::MenuItem.new( mrc['editaddr'] )
          @menuapp.add @item_editaddr
          @item_quit       = Gtk::MenuItem.new( mrc['quit'] )
          @menuapp.append @item_quit
					@item_quitf      = Gtk::MenuItem.new( mrc['quitf'] )
					@menuapp.append @item_quitf
#D#=begin
        @item_gc = Gtk::MenuItem.new( 'GC' )
        @menubar.append @item_gc
        @item_gc.signal_connect( 'activate' ) do |w|
          GC.start
        end
#D#=end


      @listbox  = Gtk::HBox.new( false, 0 )
      @paned.add1 @listbox

      @clistbox = Gtk::VBox.new( false, 0 )
			@clistadj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
			@clistscr = Gtk::VScrollbar.new( @clistadj )

      @textbox = Gtk::VBox.new( false, 0 )
      @paned.add2 @textbox

      @notice = Gtk::Label.new( 'Copyright (c) 1999 Minero Aoki' )
      @notice.set_usize( 640, 20 )
      @notice.set_alignment( 0, 0.5 )
      @textbox.pack_start @notice, false, false, 0

      @textadj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      @textscr = Gtk::VScrollbar.new( @textadj )
		@listbox.pack_start @clistscr, false, false, 0
		@listbox.pack_start @textscr, false, false, 0
    @listbox.pack_start @clistbox, true, true, 0   # !!!
      @text    = Gtk::Text.new( nil, @textadj )  # Gtk::Buffer.new
      @text.set_usize( 640, 200 )
      @text.set_editable false
      @textbox.pack_start @text, true, true, 0

      @accel = Gtk::AccelGroup.new
      @accel.attach @win
      @accel.add( ?\ , 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_ahead, 'activate' )
      @accel.add( ?j, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_ahead, 'activate' )
      @accel.add( ?k, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_back, 'activate' )
      @accel.add( ?n, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_next, 'activate' )
      @accel.add( ?p, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_prev, 'activate' )
      @accel.add( ?f, Gdk::CONTROL_MASK, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_nextpage, 'activate' )
      @accel.add( ?b, Gdk::CONTROL_MASK, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_prevpage, 'activate' )
      @accel.add( ?w, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_write, 'activate' )
      @accel.add( ?a, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_reply, 'activate' )
      @accel.add( ?e, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_reedit, 'activate' )
      @accel.add( ?f, 0, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_fwd, 'activate' )
      @accel.add( ?q, Gdk::SHIFT_MASK, Gtk::AccelGroup::ACCEL_VISIBLE,
                  @item_quit, 'activate' )


      create_pixmap
      set_callback
    end


    def create_pixmap
      @win.realize
      libdir = CORE.rc['if']['libdir'] + '/'
      @mailpm, @mailmask = Gdk::Pixmap.create_from_xpm(
        @win.window, nil, libdir + 'whitemail.xpm' )
      @mimepm, @mimemask = Gdk::Pixmap.create_from_xpm(
        @win.window, nil, libdir + 'mimemail.xpm' )
    end


    def set_callback
      @win.signal_connect( 'destroy' ) do |w|
        CORE.shutdown( true )
        true
      end

      mbox_names.each do |mbname|
        mitem = Gtk::MenuItem.new( mbname )
        mitem.signal_connect( 'activate' ) do |w|
					protect( 'move to' ){
						@current.each_selection do |idx|
							mvmail( @current.name, idx, mbname )
						end
					}
        end
        @menumove.add mitem
      end

      @excheck.signal_connect( 'toggled' ) do |w|
				protect( 'change selection mode' ) {
					@current.multiselect = ! @current.multiselect
				}
      end

      @item_trash.signal_connect( 'activate' ) do |w|
				protect( 'move to trash' ){
					c = 0
					@current.each_selection do |idx|
						c += 1
						discard_mail( @current, idx )
					end
					case c
					when 0
						CORE.notice "no mail selected"
					when 1
						;
					else
						CORE.notice "discard #{c} mails"
					end
				}
      end

      @item_update.signal_connect( 'activate' ) do |w|
				protect( 'update mbox' ){
					@current.update
				}
      end

      @item_clear.signal_connect( 'activate' ) do |w|
				protect( 'clear trash' ){
					mbox = CORE.mailset['trash']
					s = mbox.size
					s.times{|c| mbox.rm 0 }
					CORE.notice "cleared #{s} mails from trash"
				}
      end

      @item_ahead.signal_connect( 'activate' ) do |w|
				protect( 'read ahead' ){
					read_ahead
				}
      end

      @item_back.signal_connect( 'activate' ) do |w|
				protect( 'read back' ){
					read_back
				}
      end

      @item_next.signal_connect( 'activate' ) do |w|
				protect( 'view next' ){
					@current.view_next
				}
      end

      @item_prev.signal_connect( 'activate' ) do |w|
				protect( 'view prev' ){
					@current.view_prev
				}
      end

			@item_nextpage.signal_connect( 'activate' ) do |w|
			  protect( 'next page' ){
				  @clistadj.set_value @clistadj.value + @clistadj.page_size
				}
			end

			@item_prevpage.signal_connect( 'activate' ) do |w|
			  protect( 'prev page' ){
				  @clistadj.set_value @clistadj.value - @clistadj.page_size
				}
			end

      @item_write.signal_connect( 'activate' ) do |w|
				protect( 'write' ){
					CORE.edit CORE.composer.new_mail
					false
				}
      end

      @item_reply.signal_connect( 'activate' ) do |w|
				protect( 'reply' ){
					orig = @current[ @current.selected ]
					unless orig then
						CORE.notice "no mail selected to reply"
						break
					end
					CORE.edit CORE.composer.replyto( orig )
					false
				}
      end

      @item_reedit.signal_connect( 'activate' ) do |w|
				protect( 're edit' ){
					orig = @current[ @current.selected ]
					unless orig then
						CORE.notice "no mail selected to edit"
						break
					end
					CORE.edit orig
				}
      end

      @item_fwd.signal_connect( 'activate' ) do |w|
				protect( 'forward' ){
					orig = @current[ @current.selected ]
					unless orig then
						CORE.notice "no mail selected to forward"
						break
					end
					CORE.edit CORE.composer.forward( orig )
					false
				}
      end

      @item_send.signal_connect( 'activate' ) do |w|
				protect( 'send' ){
					CORE.sender.sendmail
					false
				}
      end

      @item_pop.signal_connect( 'activate' ) do |w|
				protect( 'pop' ){
        # Thread.fork do
          CORE.popper.load
        # end
					false
				}
      end

      @item_editaddr.signal_connect( 'activate' ) do |w|
				protect( 'edaddr' ){
					CORE.address_book.edit do |ret|
						if ret then
							CORE.notice 'editing address book was done.'
						else
							CORE.notice 'editing address book was canceled.'
						end
					end
					false
				}
      end

      @item_quit.signal_connect( 'activate' ) do |w|
				protect( 'shutdown' ){
					CORE.shutdown( false )
					false
				}
      end

      @item_quitf.signal_connect( 'activate' ) do |w|
				protect( 'shutdown (force)' ){
					CORE.shutdown( true )
					false
				}
      end

      @combo_pressing = false

      @mboxcombo.entry.signal_connect( 'changed' ) do |w|
				protect( 'combo change' ){
					unless @combo_pressing then
						chmbox w.get_text
					end
					false
				}
      end

      @mboxcombo.button.signal_connect( 'pressed' ) do |w|
				protect( 'combo pressed' ){
					@combo_pressing = true
					false
				}
      end

      @mboxcombo.button.signal_connect( 'released' ) do |w|
				protect( 'combo released' ){
					@combo_pressing = false
					tmp = @mboxcombo.entry.get_text
					chmbox tmp
					false
				}
      end
    end


    def nombox( name )
      nil
    end


    def realizembox( name, rc, lo )
      # box = Gtk::HBox.new( false, 0 )
      # adj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      # scr = Gtk::VScrollbar.new( adj )

      ifrc = CORE.rc['if']['list']
      cli = Gtk::CList.new([ '', '',
        ifrc['date'], ifrc['from'], ifrc['subject']])
      cli.set_usize( 640, 170 )
      cli.set_column_width( 0, 16 )
      cli.set_column_width( 1, 16 )
      cli.set_column_width( 2, 110 )
      cli.set_column_width( 3, 200 )
      # cli.set_vadjustment adj
      cli.set_vadjustment @clistadj
      cli.selection_mode = Gtk::SELECTION_BROWSE

      # box.pack_start scr, false, false, 0
      # box.pack_start cli, true, true, 0

      # @clistbox.pack_start( box, true, true, 0 )
      @clistbox.pack_start( cli, true, true, 0 )

      ret = RealMailbox.new( name, rc, lo,
        # box, cli, adj, @mailpm, @mailmask, @mimepm, @mimemask )
        cli, cli, @clistadj, @mailpm, @mailmask, @mimepm, @mimemask )

      return ret
    end

  end   # class Mailset

  

	####
	#### mbox
	####

  class Mailbox

    def initialize( name, rc, lo,
        base, cli, adj, pm_mail, msk_mail, pm_mime, msk_mime )
      super( name, rc, lo )

      @basewin = base
      @win = cli
      @adj = adj
      @mailpm   = pm_mail
      @mailmask = msk_mail
      @mimepm   = pm_mime
      @mimemask = msk_mime

      @clistsize = 0
      @selected = 0
      @unselected = false
      @multisel = false

      @win.signal_connect( 'select_row' ) do |w,row,col,eve|
        unless @multisel then
          viewmail row
        end
        false
      end

      @win.signal_connect( 'unselect_row' ) do |w,row,col,eve|
        @unselected = true
      end

			if RealMailbox === self then
			  CORE.notice 'reading mails ...'
			end
      do_initialize
    end


    attr :selected
    attr :unselected

    def select( idx )
      if idx >= 0 and idx < size then
        @selected = idx
        @win.select_row idx, 0
      end
    end

    def each_selection
      @win.each_selection{|idx| yield idx }
    end


    def multiselect
      @multisel
    end

    def multiselect=( arg )
      @multisel = arg
      if @multisel then
        @win.selection_mode = Gtk::SELECTION_EXTENDED
      else
        @win.selection_mode = Gtk::SELECTION_BROWSE
      end
    end


    def view_next
      adjust_page @selected + 1
    end

    def view_prev
      adjust_page @selected - 1
    end

    def adjust_page( idx )
      return if idx < 0 or idx >= size

      val = [ idx * @win.row_height - @adj.page_size / 2, @adj.lower ].max
      upp = [ @adj.upper - @adj.page_size, @adj.lower ].max
      @adj.value = [ val, upp ].min
      select idx
    end
    private :adjust_page

    
    def repaint
      @win.freeze
      each_index do |i|
        listmail i
      end
      @win.thaw

      select 0
    end
    

    def add( port, cache = nil )
      do_add( port )
      if cache then
        port.unread = cache[0]
        port.mime = cache[1]
        cache[0] = cache[1] = ''
        @win.append cache
        if port.mime then
          @win.set_pixmap( self.size - 1, 1, @mimepm, @mimemask )
        end
      else
        @win.append ['','','','','']
        listmail self.size - 1
      end
    end

    #def cp( port )
      
    #def mv( port )

    def rm( idx )
      do_rm( idx )
      @win.remove_row idx
    end


    def show
			@win.set_vadjustment @adj
      @basewin.show_all
    end

    def hide
      @basewin.hide_all
			@win.set_vadjustment nil
    end


    private


    def cache_entries
      all = []
      each_with_index do |port, idx|
				ent = [ port, port.unread, port.mime ]
        2.upto( @win.columns - 1 ) do |i|
          ent.push @win.get_text( idx, i )
        end
				all.push ent
      end

			all
    end

		def correct_cache?( obj )
		  true
		end

		def cache_signature
		  nil
		end


    def listmail( idx )
      port = self[idx]
      mail = Mail.new( port )
      arr = CORE.composer.list_items( mail )

      arr.each_with_index do |te,i|
        @win.set_text( idx, i + 2, te )
      end

      if port.unread then
        @win.set_pixmap( idx, 0, @mailpm, @mailmask )
      end
      if mail.multipart? then
        port.mime = true
        @win.set_pixmap( idx, 1, @mimepm, @mimemask )
      else
        port.mime = false
      end
    end


    def viewmail( idx )
      return if idx >= self.size

      @selected = idx
      return unless @win.visible?
      port = self[idx]

      CORE.notice ''
      port.unread = false
      @win.set_text( idx, 0, '' )

      CORE.mailset.view Mail.new( port )
    end

  end   # class Mailbox



  ####
	#### agent
	####

  class Popper

    def if_loggingin( client )
      @poppednum = 0
      super
      CORE.notice "connecting #{client.address}..."
    end

    def if_loggedin( client )
      CORE.notice "connected to #{client.address}"
      super
    end

    def if_popmail( port )
      super
      @poppednum += 1
      CORE.notice "pop mail \##{@poppednum}"
    end

    def if_loggedout( client )
      super
      CORE.notice "total #{@poppednum} mails popped."
    end
  
  end



  class Sender

    def if_nosendmail
      CORE.notice "no mail to send"
    end

    def if_loggedin( client )
      CORE.notice "connected to #{client.address}"
      @sentnum = 0
      super
    end

    def if_sendmail( mail, idx )
      super
      @sentnum += 1
      CORE.notice "sent mail #{@sentnum}"
    end

    def if_loggedout( client )
      super
      CORE.notice "sent total #{@sentnum} mails"
    end
  
  end



  class Composer

    def undecodable_mail( mail )
      CORE.notice "cannot decode this mail"
      return super
    end

    def unknown_type( mail )
      CORE.notice "cannot display this mail"
      return super
    end
  
  end
      


  ####
	#### editer
	####

  class Editer

    def initialize( port, rc )
      port.must Port

      @port = port
      @port.ref
      @rc = rc
      @shuttingdown = false
      do_initialize
    end

    # for ruby level exit
    def shutdown( force = false )
      return if @shuttingdown
      @shuttingdown = true
      return unless force or shutdown_ok?
      edit_cancel
    end


    private


    def edit_ok
      i = mail2idx( @port )
      CORE.mailset['draft'].select i
      CORE.notice "edit draft\##{i + 1}"
      in_shutdown
    end

    def edit_cancel
      i = mail2idx( @port )
      CORE.mailset.rmmail 'draft', i
      CORE.notice "discard draft\##{i + 1}"
      in_shutdown
    end

    def in_shutdown
      @port.unref
      CORE.edit_done self
    end


    def mail2idx( port )
      CORE.mailset['draft'].index port
    end

  end   # class Editer



  class InternalEditer < Editer

    private

    def do_initialize
      @mail = Mail.new( @port )
      @buf = SizedStack.new

      set_keybind
      construct_gui
      @win.show_all
    end


    def construct_gui
      mrc = @rc['menu']

      @win     = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest editer' )
      @win.border_width(0)
      @win.set_usize( 600, 700 )

      @box     = Gtk::VBox.new( false, 0 )
      @win.add @box

      @butbox  = Gtk::HBox.new( false, 0 )
      @box.pack_start( @butbox, false, false, 0 )

      @finbtn  = Gtk::Button.new( mrc['done'] )
      @finbtn.set_usize( 120, 22 )
      @butbox.pack_start( @finbtn, false, false, 0 )

      @intbtn  = Gtk::Button.new( mrc['cancel'] )
      @intbtn.set_usize( 120, 22 )
      @butbox.pack_start( @intbtn, false, false, 0 )

      @tobox   = Gtk::HBox.new( false, 1 )
      @box.pack_start( @tobox, false, false, 0 )

      @tolab   = Gtk::Label.new( mrc['to'] )
      @tolab.set_usize( 80, 22 )
      @tobox.pack_start( @tolab, false, false, 0 )

      @toent   = Gtk::Entry.new
      @toent.set_usize( 460, 22 )
      @toent.set_editable true
      @toent.set_text @mail.to
      @tobox.pack_start( @toent, true, true, 0 )

      @sbjbox  = Gtk::HBox.new( false, 1 )
      @box.pack_start( @sbjbox, false, false, 0 )

      @sbjlab  = Gtk::Label.new( mrc['subject'] )
      @sbjlab.set_usize( 80, 22 )
      @sbjbox.pack_start( @sbjlab, false, false, 0 )

      @sbjent  = Gtk::Entry.new
      @sbjent.set_usize( 460, 22 )
      @sbjent.set_editable true
      @sbjent.set_text @mail.subject
      @sbjbox.pack_start( @sbjent, true, true, 0 )


      @textbox = Gtk::HBox.new( false, 0 )
      @adj     = Gtk::Adjustment.new(0,0,0,0,0,0)
      @scr     = Gtk::VScrollbar.new( @adj )
      @text    = Gtk::Text.new( nil, @adj )
      # @text    = Gtk::Buffer.new( nil, @adj )
      @text.insert_text( CORE.composer.mail_to_viewtext(@mail), 0 )
      @text.set_editable true

      @textbox.pack_start( @text, true, true, 0 )
      @textbox.pack_start( @scr, false, false, 0 )
      @box.pack_start( @textbox, true, true, 0 )

      set_callback
    end


    def set_callback
      @finbtn.signal_connect( 'clicked' ) do |w|
				protect( 'edit fin' ){
					edit_ok
				}
      end

      @intbtn.signal_connect( 'clicked' ) do |w|
				protect( 'edit cancel' ){
					edit_cancel
				}
      end

      @text.signal_connect( 'key_press_event' ) do |w, eve|
        c = (eve.state & Gdk::CONTROL_MASK != 0)
        a = (eve.state & Gdk::MOD1_MASK != 0)

        bind = if    c and a then @ctrl_alt
               elsif c       then @ctrl
               elsif       a then @alt
               else               @raw
               end
#p eve.keyval
#p bind[ eve.keyval ]
        if blk = bind[ eve.keyval ] then
          emitstop = blk.call( eve )
          @text.signal_emit_stop( 'key_press_event' ) if emitstop
        end
      end
    end


    def set_keybind
      @raw = {}
      @ctrl = {}
      @alt  = {}
      @ctrl_alt = {}

      @ctrl[Gdk::GDK_i] = @ctrl[Gdk::GDK_i] = proc{
        File.open( File.expand_path '~/.signature' ) do |f|
          @text.insert_text f.read, @text.position
        end
        true
      }
      @ctrl[Gdk::GDK_k] = @ctrl[Gdk::GDK_K] = proc{
        pnt = @text.position
				@text.set_point pnt
        if @text.get_chars( pnt, pnt + 1 ) == "\n" then
					@text.forward_delete 1
          true
        else
          false
        end
      }
      @ctrl[Gdk::GDK_s] = @ctrl[Gdk::GDK_S] = proc{
        edit_ok
        false
      }
      @ctrl[Gdk::GDK_q] = @ctrl[Gdk::GDK_Q] = proc{
        edit_cancel
        false
      }
      @ctrl[Gdk::GDK_y] = @ctrl[Gdk::GDK_Y] = proc{
        if cmd = @buf.pop then
				  cmd.yank(@text)
				end
        true
      }
    end


    def edit_ok
      @mail.body    = @text.get_chars( 0, @text.get_length )
      @mail.subject = @sbjent.get_text
      @mail.to      = @toent.get_text
      @mail.sync
      super
    end
  

		def in_shutdown
      @win.hide
      @win.destroy
      super
    end


    def shutdown_ok?
      true
    end

  end


  class SizedStack

    def initialize( up = 10 )
      @upper = up
      @arr = []
    end

    attr :upper, true

    def push( item )
      @arr.push item
      if @arr.size > @upper then
        @arr.shift
      end
    end

    def pop
      @arr.pop
    end

    def each
      @arr.each{|i| yield i }
    end

    def size
      @arr.size
    end

  end


  class Backup
  end

  class DelBackup < Backup
  
    def initialize( from, str )
      @from = from
      @str = str
    end

    def undo( text )
      text.insert_text @str, @from
    end

    def yank( text )
      text.insert_text @str, text.get_point
    end
  
  end

  class MultiBackup < Backup

    def initialize( *args )
      @args = args
    end

    def undo( text )
      @args.each{|cmd| cmd.undo( text ) }
    end

    def yank( text )
      @args.each{|cmd| cmd.yank( text ) }
    end
  
  end


  class ExternalEditer < Editer

    private


    def shutdown_ok?
      true
    end


    def do_initialize
      fname = sprintf( '/tmp/est%x_%x', $$, @port.id )
      File.open( fname, 'w' ) do |f|
        @port.ropen do |rp|
          while tmp = rp.read( 2048 )
            f.write tmp
          end
        end
      end

      Thread.fork do
        do_edit( fname )
      end
    end


    def do_edit( fname )
      stamp = File.mtime( fname )
      system "#{@rc['type']} #{fname}"
      if File.mtime( fname ) == stamp then   # not canceled
        edit_cancel
      else                                   # edited
        @port.wopen do |wp|
          File.open( fname ) do |f|
            while tmp = f.read( 2048 ) do
              wp.write tmp
            end
          end
        end
        edit_ok
      end
    end

  end



  ####
	#### address
	####

  class AddressEntry
    
    def initialize( addr, name, keys = [] )
			keys.must Array
      self.addr = addr
      self.name = name
      @props = keys
    end

    attr :addr

    def addr=( adr )
      @addr = adr.dup
    end

    attr :name

    def name=( nm )
      @name = nm.dup
    end

    def properties
      @props.dup
    end

    def properties=( arr )
      arr.must Array
			arr.each {|i| i.must String }
      @props = arr.dup
    end

    def add_property( prop )
      @props |= prop
    end

    def del_property( prop )
      @props.delete prop
    end

  end


  class AddressBook

    def initialize( fname )
      @filename = File.expand_path( fname )
      @whole = {}
      @prop_tbl = {}
      File.open( @filename ) do |f|
        parse( f )
      end

      @edit = @get_addr = nil
    end


    def parse( f )
      in_group = false

      while line = f.gets do
        line.strip!
        next if line == '%'

        addr = line
        line = f.gets ; line.chomp! ; line.strip! ; name = line
        line = f.gets ; line.chomp! ; line.strip! ; keys = line.split(',')

        ent = AddressEntry.new( addr, name, keys )
        @whole[ addr ] = ent
        keys.each do |k|
          k.strip!
          next if k.size == 0
          unless arr = @prop_tbl[k] then
            @prop_tbl[k] = arr = []
          end
          arr.push ent
        end
      end
    end


    def get_addr( &block )
      unless @get_addr then
        @get_addr = GetAddress.new( self )
      end
      return @get_addr.invoke( block )
    end


    def edit( &block )
      unless @edit then
        @edit = EditAddress.new( self )
      end
      return @edit.invoke( block )
    end

    
    def []( addr )
      return @whole[ addr ]
    end

    def store( addr, entry )
      if @whole.has_key? addr then
        raise ArgumentError, "same address '#{addr}' was registered"
      end
      @whole[addr] = entry
    end

    def each
      @whole.each{|k,ent| yield ent }
    end

    def values
      @whole.values
    end

    def keys
      @whole.keys
    end

    def properties
      a = @prop_tbl.keys
			a.sort!
			a
    end

    def new_property( prop )
      if @prop_tbl.has_key? prop then
        raise ArgumentError, "property '#{prop}' is already registered"
      end
      @prop_tbl[ prop ] = []
    end

    def lookup( keys )
      ret = {}
      keys.each do |k|
        if arr = @prop_tbl[k] then
          arr.each{|a| ret[a] = true }
        end
      end
      return ret.keys
    end

  end


  class ModalTask

    def initialize( book )
      @pipe = Queue.new
      @book = book
      construct_gui
    end

    def invoke( block )
      @win.show_all
      Thread.fork do
        ret = @pipe.pop
        @win.hide_all
        if ret then
          block.call @work.call
        else
          block.call nil
        end
      end
    end

    private

    def task_end( result )
      @pipe.push result
    end

  end


  class EditAddress < ModalTask

    def invoke( block )
      reset
      super
    end


    private


    def strarr2list( arr )
      arr.filter do |i|
			  li = Gtk::ListItem.new( i )
				li.show
				li
			end
      arr
    end

    def reset
      @list.clear
      entries = @book.values.sort{|a,b| a.name <=> b.name }
      entries.each_with_index do |ent, idx|
        @list.append [ent.name, ent.properties.join(','), ent.addr]
        @list.set_row_data idx, ent
      end

      @proplist.clear_items( 0, -1 )

      @allproplist.clear_items( 0, -1 )
      @allproplist.append_items strarr2list( @book.properties.sort )
      
      if ent = entries[0] then
        @cur_row = 0
        @cur_ent = ent
        write_in
      end
    end


    def write_in
      @name_entry.set_text @cur_ent.name
      @address_entry.set_text @cur_ent.addr
      @proplist.clear_items( 0, -1 )
      @proplist.append_items strarr2list( @cur_ent.properties )
    end

    def write_back
      @cur_ent.name = nm = @name_entry.get_text
      @cur_ent.addr = ad = @address_entry.get_text

			arr = @cur_ent.properties
      @list.set_text @cur_row, 0, nm
      @list.set_text @cur_row, 1, arr.join(',')
      @list.set_text @cur_row, 2, ad
    end


    def set_callback
      @win.signal_connect( 'destroy' ) do |w|
        task_end false
      end

      @list.signal_connect( 'select_row' ) do |w, row, col, eve|
        write_back
        @cur_row = row
        @cur_ent = @list.get_row_data( row )
        write_in
      end

			@del_button.signal_connect( 'clicked' ) do |w|
			  sel = @proplist.selection
				unless sel.empty? then
					sel.each do |i|
					  @cur_ent.del_property i.child.get_text
					end
					@proplist.remove_items sel
				end
			end

			@add_button.signal_connect( 'clicked' ) do |w|
			  sel = @allproplist.selection
				unless sel.empty? then
					sel.each do |i|
					  @cur_ent.add_property i.child.get_text
					end
					@proplist.clear_items 0, -1
					@proplist.append_items strarr2list( @cur_ent.properties )
				end
			end

      @ok_button.signal_connect( 'clicked' ) do |w|
        task_end false #true
        false
      end

      @cancel_button.signal_connect( 'clicked' ) do |w|
        task_end false
        false
      end
    end


    def construct_gui
      @win = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_modal true
			@win.set_title 'mailest address book'

      @box = Gtk::VBox.new( false, 0 )
      @win.add @box

      @message = Gtk::Label.new( 'choose data to edit' )
			@message.set_alignment 0, 0.5
      @box.pack_start @message, false, false, 5

      @listbox = Gtk::HBox.new( false, 0 )
      @box.pack_start @listbox, true, true, 0

      @list = Gtk::CList.new( ['name', 'property', 'address'] )
      @list.set_column_width 0, 150
      @list.set_column_width 1, 300
      @list.set_column_width 2, 150
      @listbox.pack_start @list, true, true, 0

      @listvadj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      @listvscr = Gtk::VScrollbar.new( @listvadj )
      @list.vadjustment = @listvadj
      @listbox.pack_start @listvscr, false, false, 0

      @name_entry = Gtk::Entry.new
      @name_entry.set_editable true
			@box.pack_start @name_entry, false, false, 0
        
      @address_entry = Gtk::Entry.new
      @address_entry.set_editable true
			@box.pack_start @address_entry, false, false, 0

			@bottom_box = Gtk::HBox.new( false, 0 )
			@box.pack_start @bottom_box, true, true, 0


			@plist_box = Gtk::VBox.new( false, 0 )
			@bottom_box.pack_start @plist_box, true, true, 0
			@plist_box_in = Gtk::HBox.new( false, 0 )
			@plist_box.pack_start @plist_box_in, true, true, 0

			@allplist_box = Gtk::VBox.new( false, 0 )
			@bottom_box.pack_start @allplist_box, true, true, 0
			@allplist_box_in = Gtk::HBox.new( false, 0 )
			@allplist_box.pack_start @allplist_box_in, true, true, 0

      @proplist = Gtk::List.new
			@proplist.set_selection_mode Gtk::SELECTION_BROWSE
			@plist_box_in.pack_start @proplist, true, true, 0
			@proplist_scr = Gtk::VScrollbar.new
			@plist_box_in.pack_start @proplist_scr, false, false, 0
			@del_button = Gtk::Button.new( 'del' )
			@plist_box.pack_start @del_button, false, false, 0

      @allproplist = Gtk::List.new
			@allproplist.set_selection_mode Gtk::SELECTION_BROWSE
			@allplist_box_in.pack_start @allproplist, true, true, 0
			@allproplist_scr = Gtk::VScrollbar.new
			@allplist_box_in.pack_start @allproplist_scr, false, false, 0
			@add_button = Gtk::Button.new( 'add' )
			@allplist_box.pack_start @add_button, false, false, 0


			@button_box = Gtk::VBox.new( true, 0 )
			@bottom_box.pack_start @button_box, false, false, 0

      @ok_button = Gtk::Button.new( 'edit done' )
      @button_box.pack_start @ok_button, false, false, 2

      @cancel_button = Gtk::Button.new( 'cancel' )
      @button_box.pack_start @cancel_button, false, false, 2

      set_callback
    end

  end


  class GetAddress < ModalTask

    def invoke( block )
      @to = []
      @cc = []
      @bcc = []
      return super
    end


    def construct_gui
      @win = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_modal true
      @win.signal_connect( 'destroy' ) do |w|
        task_end false
      end

      @box = Gtk::HBox.new( false, 0 )
      @win.add @box

      @list = Gtk::CList.new( ['', 'name', 'property'] )
      @list.set_column_width 0, 50
      @list.set_column_width 1, 150
      @list.set_column_width 2, 300

      @book.values.sort{|a,b| a.name <=> b.name }.each do |ent|
        @list.append ['', ent.name, ent.keys.join(',')]
        @list.set_row_data @list.rows - 1, ent.addr
      end

      @list.signal_connect( 'select_row' ) do |w, row, col, eve|
        ent = @list.get_row_data( row )

        sel = @to
        pix = 'To' #@to_pix
        if    @to_button.active? then  sel = @to  ; pix = 'To'
        elsif @cc_button.active? then  sel = @cc  ; pix = 'CC'
        elsif @bcc_button.active? then sel = @bcc ; pix = 'BCC'
        end

        sel.push ent
        [@to, @cc, @bcc].each {|arr| arr.delete ent unless arr == sel }
        # @list.set_pixmap row, 0, pix
        @list.set_text row, 0, pix
      end

      @list.signal_connect( 'unselect_row' ) do |w, row, col, eve|
        @list.set_text row, 0, ''
        ent = @list.get_row_data( row )
        [@to, @cc, @bcc].each do |arr|
          if arr.delete ent then
            break
          end
        end
      end

      @box.pack_start @list, false, false, 0


      @console_box = Gtk::VBox.new( false, 0 )
      @box.pack_start @console_box

      @message = Gtk::Label.new( 'select To/Cc/Bcc' )
      @console_box.pack_start @message, false, false, 5

      @to_button  = Gtk::RadioButton.new( nil, 'To' )
      @cc_button  = Gtk::RadioButton.new( @to_button.group, 'Cc' )
      @bcc_button = Gtk::RadioButton.new( @to_button.group, 'Bcc' )
      @console_box.pack_start @to_button,  false, false, 3
      @console_box.pack_start @cc_button,  false, false, 3
      @console_box.pack_start @bcc_button, false, false, 3

      @ok_button = Gtk::Button.new( 'Selection OK' )
      @ok_button.signal_connect( 'clicked' ) do |w|
        task_end false #true
        false
      end
      @console_box.pack_start @ok_button, false, false, 2

      @cancel_button = Gtk::Button.new( 'Cancel to write a mail' )
      @cancel_button.signal_connect( 'clicked' ) do |w|
        task_end false
        false
      end
      @console_box.pack_start @cancel_button, false, false, 2
    end

    private :construct_gui

  end


end   # module EST


class Port

  attr :mime, true

end
