require 'racc/parser'
require 'racc/scanner'
require 'amstd/recycle'

#
# mailp.rb
#
#     Copyright(c) 1999 Minero Aoki
#     aamine@dp.u-netsurf.ne.jp
#

$mailpdebugflag = false


class Mailp < Parser
  include Recyclable

  def initialize
    @scanner = MailScanner.new
  end


  def reset
    self.debug = $mailpdebugflag
  end


  def debug=( flag )
    @yydebug = flag
    @scanner.debug = @yydebug && DEBUG_PARSER
  end


  def debug
    @yydebug
  end


  def Mailp.parse( hder, str )
    use_instance do |inst|
      inst.parse( hder, str )
    end
  end


  def parse( hder, str )
    hder.must HeaderField
    str.must String

    if /\A\s+\z/ === str then return end
    str = str.dup

    @field  = hder

    case hder
    when MsgidH
      msgid_parse( str )
    when RefH
      refs_parse( str )
    else
      @scanner.reset( str, @field )

      sim = hder.type.to_s
      sim = sim.split('::')[-1].to_s.intern
      @cache = [sim, nil]

      do_parse
    end
  end


  private


  def next_token
    if ret = @cache then
      @cache = nil
    else
      ret = @scanner.scan
    end
    ret
  end


=begin c
  def next_value
    if @valarr.size == 0 then
      @scanner.scan
    end
    return @valarr.shift
  end


  def peep_token
    if @simarr.size == 0 then
      ret = @scanner.scan( false )
    else
      ret = @simarr[0]
    end

    return ret
  end
=end c

  
  def on_error( tok, val, _tstack, _vstack, _state )
    raise ParseError,
      "\nparse error in '#{@field.name}' header, on token #{val.inspect}"
  end


  def refs_parse( str )
    temp = []

    while MsgidH::MSGID === str do
      temp.push( ($`).strip )
      temp.push( $& )
      str = $'
    end
    temp.push str

    @field.refs.replace temp
  end


  def msgid_parse( str )
    if MsgidH::MSGID === str then
      @field.msgid = $&
    else
      raise ParseError,
        "wrong Message-ID format: #{str}"
    end
  end



  class MailScanner < Scanner

    # mode

    MODE_NONE = :MODE_NONE

    MODE_822  = :MODE_822
    MODE_MIME = :MODE_MIME

    MODE_RECV = :MODE_RECV


    def reset( str, fld )
      super( str )
      @header = fld

      @m_atom = MODE_822
      @m_recv = MODE_NONE

      case fld
      when CTypeH
        @m_atom = MODE_MIME
      when CEncodingH, CDispositionH
        @m_atom = MODE_MIME
      when RecvH
        @m_recv = MODE_RECV
      end
    end


    # exp

    atomchars  = Regexp.quote( "\#!$%&`'*+{|}~^/=?" ) + '\\-'
    tokenchars = Regexp.quote( "\#!$%&`'*+{|}~^." )   + '\\-'
    eucchars   = "\xa1\xa1-\xf3\xfe"
    jisstr     = "\\e..[^\\e]*\\e.."

    ATOM    = /\A[\w#{atomchars}#{eucchars}]+|#{jisstr}/o
    TOKEN   = /\A[\w#{tokenchars}#{eucchars}]+|#{jisstr}/o
    DIGIT   = /\A\d+\z/o

    LWSP    = /\A(?:\n|\r\n|\r)?[ \t]+/o

    BACKSLASH = /\A\\/o

    BEGIN_Q = /\A"/o
    Q_ENT   = /\A[^"\\\e]+|#{jisstr}/o
    END_Q   = /\A"/o

    BEGIN_C = /\A\(/o
    C_ENT   = /\A[^\)\(\\\e]+|#{jisstr}/o
    END_C   = /\A\)/o

    BEGIN_D = /\A\[/o
    D_ENT   = /\A[^\]\\]+|#{jisstr}/o
    END_D   = /\A\]/o


    OMIT = '$omit'

    def scan( cut = true )
      sret = nil
      vret = OMIT

      until sret do
        unless @scan.rest? then
          sret = false; vret = false
          break
        end

        @scan.skip LWSP

        if @m_atom == MODE_822 then
          if vret = @scan.fullscan( ATOM, true, cut ) then
            sret = :ATOM

            if DIGIT === vret then
              sret = :DIGIT
              vret = vret

            elsif @m_recv = MODE_RECV then
              case vret.downcase
              when 'from' then sret = :FROM
              when 'by'   then sret = :BY
              when 'via'  then sret = :VIA
              when 'with' then sret = :WITH
              when 'id'   then sret = :ID
              when 'for'  then sret = :FOR
              end
            end
            break
          end

        elsif @m_atom == MODE_MIME then
          if vret = @scan.fullscan( TOKEN, cut, cut ) then
            sret = :TOKEN
            break
          end
        else
          bug! 'atom mode is not 822/MIME'
        end

        if @scan.fullscan( BEGIN_Q, cut, cut ) then
          sret = :QUOTED
          vret = quoted if cut

        elsif @scan.skip( BEGIN_C ) then
          @header.comments.push comment

        elsif @scan.fullscan( BEGIN_D, cut, cut ) then
          sret = :DOMLIT
          vret = domlit if cut

        else
          sret = vret = @scan.getch
          unless cut then
            @scan.unscan
          end
        end
      end

      ret = [sret, vret]
      debug_report( ret ) if @debug
      ret
    end



    private


    def quoted
      ret = ''
      while true do
        if    temp = @scan.scan( Q_ENT ) then ret << temp
        elsif @scan.skip( END_Q )        then break
        elsif @scan.skip( BACKSLASH )    then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated quoted-string"
          end
          bug! 'in quoted, no match'
        end
      end

      return ret
    end

    
    def comment
      ret = ''
      nest = 1

      while nest > 0 and @scan.rest? do
        if    temp = @scan.scan( C_ENT ) then ret << temp
        elsif @scan.skip( END_C )        then nest -= 1
        elsif @scan.skip( BEGIN_C )      then nest += 1
        elsif @scan.skip( BACKSLASH )    then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated comment"
          end
          bug! 'in comment, no match'
        end
      end
      if nest > 0 then
        scan_error! "found unterminated comment"
      end

      return ret
    end

    
    def domlit
      ret = ''

      while true do
        if temp = @scan.scan( D_ENT ) then
          ret << temp
        end
        if    @scan.skip( END_D )     then break
        elsif @scan.skip( BACKSLASH ) then ret << @scan.getch
        else
          unless @scan.rest? then
            scan_error! "found unterminated domain literal"
          end
          bug! 'in domlit, no match'
        end
      end

      return ret
    end


    def report_ret( sret, ret )
      if Integer === sret then
        puts "sret #{sret.id2name || sret}"
      else
        puts "sret #{sret.inspect}"
      end
      puts "vret #{ret}"
    end

  end

##### racc generated code begin ###


LR_shift_n = 162

LR_reduce_table = [0, 0, :racc_error,

2, 4, :_reduce_1,
2, 4, :_reduce_2,
2, 4, :_reduce_3,
2, 4, :_reduce_4,
2, 4, :_reduce_5,
2, 4, :_reduce_6,
2, 4, :_reduce_7,
2, 4, :_reduce_8,
2, 4, :_reduce_9,
2, 4, :_reduce_10,
2, 4, :_reduce_11,
2, 4, :_reduce_12,
2, 4, :_reduce_13,
2, 4, :_reduce_14,
2, 4, :_reduce_15,
6, 3, :_reduce_16,
0, 33, :_reduce_17,
2, 33, :_reduce_18,
3, 36, :_reduce_19,
5, 36, :_reduce_20,
1, 37, :_reduce_21,
7, 6, :_reduce_22,
0, 40, :_reduce_23,
2, 40, :_reduce_24,
4, 40, :_reduce_25,
3, 40, :_reduce_26,
0, 41, :_reduce_27,
2, 41, :_reduce_28,
0, 42, :_reduce_29,
2, 42, :_reduce_30,
0, 43, :_reduce_31,
3, 43, :_reduce_32,
0, 44, :_reduce_33,
2, 44, :_reduce_34,
2, 44, :_reduce_35,
0, 45, :_reduce_36,
2, 45, :_reduce_37,
0, 46, :_reduce_38,
2, 46, :_reduce_39,
2, 8, :_reduce_40,
1, 8, :_reduce_41,
1, 10, :_reduce_42,
3, 10, :_reduce_43,
1, 12, :_reduce_44,
1, 12, :_reduce_45,
1, 14, :_reduce_46,
3, 14, :_reduce_47,
1, 16, :_reduce_48,
1, 16, :_reduce_49,
2, 16, :_reduce_50,
4, 60, :_reduce_51,
4, 59, :_reduce_52,
3, 59, :_reduce_53,
2, 63, :_reduce_54,
2, 64, :_reduce_55,
4, 64, :_reduce_56,
3, 61, :_reduce_57,
1, 61, :_reduce_58,
1, 65, :_reduce_59,
3, 65, :_reduce_60,
1, 48, :_reduce_61,
3, 48, :_reduce_62,
1, 68, :_reduce_63,
1, 68, :_reduce_64,
1, 68, :_reduce_65,
3, 18, :_reduce_66,
1, 62, :_reduce_67,
2, 62, :_reduce_68,
1, 66, :_reduce_69,
1, 66, :_reduce_70,
1, 66, :_reduce_71,
1, 22, :_reduce_72,
3, 22, :_reduce_73,
1, 24, :_reduce_74,
2, 24, :_reduce_75,
3, 26, :_reduce_76,
4, 28, :_reduce_77,
2, 28, :_reduce_78,
0, 73, :_reduce_79,
5, 73, :_reduce_80,
1, 75, :_reduce_81,
1, 75, :_reduce_82,
1, 30, :_reduce_83,
2, 32, :_reduce_84,
0, 76, :_reduce_85,
3, 76, :_reduce_86,
0, 77, :_reduce_87,
3, 77, :_reduce_88,
3, 77, :_reduce_89,
3, 77, :_reduce_90,
1, 69, :_reduce_91,
1, 69, :_reduce_92,
1, 69, :_reduce_93,
1, 69, :_reduce_94,
1, 69, :_reduce_95,
1, 69, :_reduce_96,
1, 69, :_reduce_97 ]

LR_reduce_n = 98

LR_action_table = [
     5,     1,    19,     3,    31,     2,    15,     5,    27,     4,
    11,     7,    23,     6,     7,     8,     2,    10,    21,    11,
    17,    13,    29,    12,    13,    15,    25,    14,     9,    16,
     0,   163,    47,    17,     0,   -23,    71,    20,     0,   163,
    20,    22,     0,   163,    71,    23,     0,   163,    47,    25,
    55,    26,    35,    28,    51,    27,    70,    31,    54,    30,
    57,    32,    34,    35,    53,    34,    52,    40,     0,   163,
    47,    25,    55,    26,    35,    28,    51,    27,    70,    31,
    54,    30,    34,    35,    53,    34,    52,    40,     0,   163,
    47,    25,    55,    26,    35,    28,    51,    27,    70,    31,
    54,    30,    57,    32,    34,    35,    53,    34,    52,    40,
     0,   163,    57,    49,     0,   163,    -1,    51,     0,   163,
    35,    52,    34,   -17,     0,   163,    47,    25,    55,    26,
    35,    28,    51,    27,    70,    31,    54,    30,    34,    35,
    53,    34,    52,    40,     0,   163,    71,    59,     0,   163,
    57,    61,     0,   163,    34,    63,     0,   163,    47,    25,
    55,    26,    35,    28,    51,    27,    70,    31,    54,    30,
    57,    32,    34,    35,    53,    34,    52,    40,     0,   163,
    47,    25,    55,    26,    35,    28,    51,    27,    70,    31,
    54,    30,    57,    32,    34,    35,    53,    34,    52,    40,
     0,   163,    47,    25,    55,    26,    35,    28,    51,    27,
    54,    30,    50,    70,    34,    72,    53,    34,    52,    40,
     0,   163,     0,    -2,    51,    74,     0,   -27,     0,   -85,
     0,   -15,     0,    -9,    72,    76,     0,   -79,     0,   -13,
     0,   -92,     0,   -97,     0,   -93,     0,   -91,     0,   -69,
     0,   -96,     0,   -70,    47,    25,    55,    26,    35,    28,
    51,    27,    70,    31,    54,    30,    34,    35,    53,    34,
    49,    81,    52,    40,     0,   163,     0,   -49,     0,   -95,
     0,   -71,    47,    25,    55,    26,    35,    28,    51,    27,
    70,    31,    54,    30,    57,    32,    34,    35,    53,    34,
    52,    40,     0,   163,     0,    -7,    38,   -59,    -1,   -59,
    67,   -59,    56,   -59,    58,   -59,    49,   -59,     0,   -67,
    67,    85,    49,    86,     0,   -58,     0,   -94,     0,   -48,
     0,   -11,    47,    25,    55,    26,    -1,   -74,    35,    28,
    51,    27,    70,    31,    54,    30,    34,    35,    53,    34,
    52,    40,     0,   163,     0,   -45,    47,    25,    55,    26,
    35,    28,    51,    27,    39,    88,    70,    31,    54,    30,
    57,    32,    34,    35,    53,    34,    52,    40,     0,   163,
     0,   -44,     0,    -5,     0,   -41,    47,    25,    55,    26,
    35,    28,    51,    27,    70,    31,    54,    30,    58,    89,
    34,    35,    53,    34,    49,    81,    52,    40,     0,   163,
     0,    -3,    -1,    90,     0,   163,    38,    91,     0,   163,
    34,    92,     0,   163,     0,    -1,    38,    93,    -1,   -10,
     0,   163,    47,    25,    55,    26,    35,    28,    51,    27,
    70,    31,    54,    30,    34,    35,    53,    34,    52,    40,
     0,   -72,     0,   -67,     0,   -14,     0,   -83,     0,    -8,
    47,    25,    55,    26,    35,    28,    51,    27,    70,    31,
    54,    30,    34,    35,    53,    34,    52,    40,     0,   163,
     0,   -12,    67,    95,     0,   163,    38,    96,    -1,    -6,
     0,   163,     0,   -46,    38,    97,    -1,    -4,     0,   163,
     0,   -42,     0,   -63,    67,    98,    50,    99,    49,   100,
     0,   -24,     0,   -64,     0,   -61,     0,   -65,    52,   102,
     0,   -29,    47,    25,    55,    26,    35,    28,    51,    27,
    54,    30,    50,    70,    34,    72,    53,    34,    52,    40,
     0,   163,    -1,   -84,    56,   104,     0,   163,    71,   105,
     0,   163,    -1,   -78,    56,   106,     0,   163,    38,   107,
    39,   108,     0,   163,    47,    25,    55,    26,    35,    28,
    51,    27,    70,    31,    54,    30,    34,    35,    53,    34,
    52,    40,     0,   163,     0,   -59,    47,    25,    55,    26,
    35,    28,    51,    27,    54,    30,    50,    70,    34,    72,
    53,    34,    52,    40,     0,   163,    58,   111,     0,   163,
     0,   -50,     0,   -68,    47,    25,    55,    26,    35,    28,
    51,    27,    70,    31,    54,    30,    34,    35,    53,    34,
    52,    40,     0,   163,    47,    25,    55,    26,    35,    28,
    51,    27,    54,    30,    50,    70,    34,    72,    53,    34,
    52,    40,     0,   163,     0,   -75,    47,    25,    55,    26,
    35,    28,    51,    27,    70,    31,    54,    30,    57,    32,
    34,    35,    53,    34,    52,    40,     0,   163,     0,   -40,
     0,   -99,     0,   -18,    35,   115,     0,   163,    47,    25,
    55,    26,    35,    28,    51,    27,    70,    31,    54,    30,
    34,    35,    53,    34,    52,    40,     0,   163,    58,   117,
     0,   163,    34,   118,     0,   163,    47,    25,    55,    26,
    35,    28,    51,    27,    70,    31,    54,    30,    57,    32,
    34,    35,    53,    34,    52,    40,     0,   163,    47,    25,
    55,    26,    35,    28,    51,    27,    70,    31,    54,    30,
    57,    32,    34,    35,    53,    34,    52,    40,     0,   163,
    47,    25,    55,    26,    35,    28,    51,    27,    54,    30,
    50,    70,    34,    72,    53,    34,    52,    40,     0,   163,
     0,   -26,    47,    25,    55,    26,    35,    28,    51,    27,
    54,    30,    50,    70,    34,    72,    53,    34,    52,    40,
     0,   163,     0,   -31,    35,   124,     0,   163,    67,    98,
     0,   -28,    71,   126,     0,   -87,     0,   -79,    71,   128,
     0,   163,    49,   129,     0,   163,     0,   -54,    58,   130,
     0,   163,    67,    98,     0,   -55,     0,   -53,     0,   -60,
    67,    98,     0,   -57,    38,    96,    56,   131,     0,   163,
    34,   132,     0,   163,    47,    25,    55,    26,    35,    28,
    51,    27,    70,    31,    54,    30,    34,    35,    53,    34,
    52,    40,     0,   -73,     0,   -66,     0,   -76,     0,   -47,
     0,   -43,     0,   -62,    67,    98,     0,   -25,    54,   133,
    53,   134,     0,   -33,     0,   -30,     0,   -86,    74,   136,
     0,   163,    -1,   -77,    56,   106,     0,   163,    74,   137,
     0,   163,    47,    25,    55,    26,    35,    28,    51,    27,
    54,    30,    50,    70,    34,    72,    53,    34,    52,    40,
     0,   163,     0,   -52,     0,   -51,    34,   140,     0,   163,
    35,   142,    57,    61,     0,   163,    35,   143,     0,   163,
    55,   144,     0,   -36,    47,    25,    55,    26,    35,    28,
    51,    27,    78,   147,    70,   148,    54,    30,    53,    34,
    52,    40,     0,   163,    70,   150,    71,   151,     0,   163,
    67,    98,     0,   -56,    35,   152,     0,   163,    39,   154,
     0,   163,     0,   -34,     0,   -35,     0,   -32,    47,    25,
    55,    26,    35,    28,    51,    27,    70,    31,    54,    30,
    57,    32,    34,    35,    53,    34,    52,    40,     0,   163,
    -1,   -38,    56,   156,     0,   163,     0,   -90,     0,   -89,
     0,   -88,     0,   -80,     0,   -82,     0,   -81,     0,   -21,
     0,   -16,    34,   158,     0,   163,     0,   -37,    35,    52,
    34,   -17,     0,   163,     0,   -22,    35,   -19,    39,   160,
     0,   163,     0,   -39,    34,   161,     0,   163,     0,   -20 ]

LR_action_table_ptr = [
     0,    32,    36,    40,    44,    48,    70,    90,   112,   116,
   120,   126,   146,   150,   154,   158,   180,   202,   222,   224,
   228,   230,   232,   234,   238,   240,   242,   244,   246,   248,
   250,   252,   254,   276,   278,   280,   282,   304,   306,   320,
   326,   328,   330,   332,   354,   356,   380,   382,   384,   386,
   410,   412,   416,   420,   424,   426,   432,   452,   454,   456,
   458,   460,   480,   482,   486,   492,   494,   500,   502,   504,
   512,   514,   516,   518,   522,   542,   548,   552,   558,   564,
   584,   586,   606,   610,   612,   614,   634,   654,   656,   678,
   680,   682,   684,   688,   708,   712,   716,   738,   760,   780,
   782,   802,   804,   808,   812,   816,   818,   822,   826,   828,
   832,   836,   838,   840,   844,   850,   854,   874,   876,   878,
   880,   882,   884,   888,   894,   896,   898,   902,   908,   912,
   932,   934,   936,   940,   946,   950,   954,   974,   980,   984,
   988,   992,   994,   996,   998,  1020,  1026,  1028,  1030,  1032,
  1034,  1036,  1038,  1040,  1042,  1046,  1048,  1054,  1056,  1062,
  1064,  1068 ]


LR_goto_table = [
     4,     9,     6,    18,    40,    19,    32,    21,    28,    24,
    69,    29,    59,    33,    62,    36,    16,    37,    66,    38,
    65,    39,    61,    41,    69,    29,    24,    42,    66,    43,
    60,    44,    69,    29,    59,    33,    62,    45,    16,    46,
    12,    47,    66,    38,    65,    39,    61,    41,    59,    48,
     8,    50,    33,    53,     3,    54,    69,    29,    22,    55,
    62,    56,    66,    57,    30,    58,    18,    60,    26,    62,
    14,    64,    69,    29,    59,    33,    62,    36,    16,    65,
    66,    38,    65,    39,    61,    41,    60,    44,    69,    29,
    10,    66,    59,    33,    62,    45,    16,    46,    12,    67,
    66,    38,    65,    39,    61,    41,    69,    68,    48,    69,
    68,    71,    41,    73,    76,    75,    73,    77,    64,    78,
    69,    29,    63,    79,    66,    80,    65,    39,    61,    82,
    69,    29,    59,    83,    66,    84,    69,    29,    66,    87,
    69,    29,    59,    83,    66,    84,    64,    78,    69,    29,
    63,    79,    66,    80,    65,    39,    61,    82,    69,    29,
    66,    84,    69,    29,    66,    80,    65,    39,    61,    94,
    42,   101,    69,    68,    48,   103,    68,    71,    69,    29,
    66,    80,    65,    39,    61,   109,    69,    68,    48,   110,
    68,    71,    69,    29,    66,   112,    69,    68,    48,   113,
    68,    71,    14,   114,    69,    29,    59,    33,    62,    36,
    16,    65,    66,    38,    65,    39,    61,    41,    69,    29,
    62,   116,    66,    57,    69,    29,    59,    33,    62,    36,
    16,   119,    66,    38,    65,    39,    61,    41,    60,    44,
    69,    29,    59,    33,    62,    45,    16,    46,    12,   120,
    66,    38,    65,    39,    61,    41,    69,    68,    68,   121,
    69,    68,    48,   122,    68,    71,    43,   123,    77,   125,
    73,   127,    69,    29,    66,    84,    44,   135,    69,    68,
    48,   138,    68,    71,    36,   139,    18,   141,    45,   145,
    69,   146,    75,   149,    37,   153,    60,    44,    69,    29,
    59,    33,    62,    45,    16,    46,    12,   155,    66,    38,
    65,    39,    61,    41,    46,   157,    33,    53,     3,   159 ]

LR_goto_table_ptr = [
     0,     2,     6,    -1,     8,    10,    24,    30,    48,    -1,
    52,    56,    64,    66,    68,    70,    86,   106,    -1,   112,
   114,    -1,    -1,   116,    -1,    -1,    -1,    -1,    -1,    -1,
    -1,    -1,   118,    -1,    -1,    -1,   130,    -1,    -1,    -1,
    -1,    -1,    -1,   136,    -1,   140,    -1,    -1,    -1,   146,
    -1,    -1,    -1,    -1,    -1,    -1,   158,    -1,    -1,    -1,
    -1,   162,    -1,    -1,    -1,    -1,    -1,    -1,    -1,    -1,
    -1,    -1,    -1,   170,   172,    -1,    -1,    -1,    -1,   178,
    -1,   186,    -1,    -1,    -1,   192,   196,    -1,   202,    -1,
    -1,    -1,    -1,   218,    -1,    -1,   224,   238,   256,    -1,
   260,   266,    -1,    -1,   268,   270,    -1,    -1,    -1,    -1,
    -1,    -1,    -1,    -1,    -1,    -1,   272,    -1,    -1,    -1,
    -1,    -1,    -1,   276,    -1,    -1,    -1,    -1,    -1,   278,
    -1,    -1,   284,   286,    -1,   288,   290,   292,    -1,   294,
    -1,    -1,    -1,    -1,   296,   314,    -1,    -1,    -1,    -1,
    -1,    -1,    -1,    -1,    -1,    -1,   316,    -1,    -1,    -1,
    -1,    -1 ]


LR_token_table = {
 :$default => 0,
 :DateH => 2,
 :RecvH => 5,
 :RetpathH => 7,
 :MaddrH => 9,
 :SaddrH => 11,
 :MmboxH => 13,
 :SmboxH => 15,
 :MsgidH => 17,
 :RefH => 19,
 :refs => 20,
 :KeyH => 21,
 :EncH => 23,
 :VersionH => 25,
 :CTypeH => 27,
 :CEncodingH => 29,
 :CDispositionH => 31,
 :DIGIT => 34,
 :ATOM => 35,
 "," => 38,
 ":" => 39,
 :FROM => 47,
 "@" => 49,
 :DOMLIT => 50,
 :BY => 51,
 :VIA => 52,
 :WITH => 53,
 :ID => 54,
 :FOR => 55,
 ";" => 56,
 "<" => 57,
 ">" => 58,
 "." => 67,
 :QUOTED => 70,
 :TOKEN => 71,
 "/" => 72,
 "=" => 74,
 :VALUE => 78,
 false => -1 }

DEBUG_PARSER = false

def _reduce_0( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_1( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.date = val[1]
 result
end

def _reduce_2( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_3( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_4( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.addrs.replace val[1]
 result
end

def _reduce_5( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.addr = val[1]
 result
end

def _reduce_6( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.addrs.replace val[1]
 result
end

def _reduce_7( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.addr = val[1]
 result
end

def _reduce_8( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.msgid = val[1]
 result
end

def _reduce_9( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.refs.replace val[1]
 result
end

def _reduce_10( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.keys.replace val[1]
 result
end

def _reduce_11( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_12( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_13( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_14( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_15( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_16( tok, val, _tokens, _values, _states )
 result = val[0]
      t = Time.gm( val[3].to_i, val[2], val[1].to_i, 0, 0, 0 )
      result = (t + val[4] - val[5]).localtime
 result
end

def _reduce_17( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_18( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_19( tok, val, _tokens, _values, _states )
 result = val[0]
 result = (val[0].to_i * 60 * 60) + (val[2].to_i * 60)
 result
end

def _reduce_20( tok, val, _tokens, _values, _states )
 result = val[0]
 result = (val[0].to_i * 60 * 60) + (val[2].to_i * 60) + val[4].to_i
 result
end

def _reduce_21( tok, val, _tokens, _values, _states )
 result = val[0]
 result = DateH.zone_s2i( val[0] ) * 60
 result
end

def _reduce_22( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.from  = val[0]
        @field.by    = val[1]
        @field.via   = val[2]
        @field.with.replace val[3]
        @field.msgid = val[4]
        @field.ford  = val[5]
        @field.date  = val[6]
 result
end

def _reduce_23( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_24( tok, val, _tokens, _values, _states )
 result = val[0]
      result = val[1]
 result
end

def _reduce_25( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[3]
 result
end

def _reduce_26( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[1]
 result
end

def _reduce_27( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_28( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[1]
 result
end

def _reduce_29( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_30( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[1]
 result
end

def _reduce_31( tok, val, _tokens, _values, _states )
 result = val[0]
        result = []
 result
end

def _reduce_32( tok, val, _tokens, _values, _states )
 result = val[0]
        result.push val[2]
 result
end

def _reduce_33( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_34( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[1]
 result
end

def _reduce_35( tok, val, _tokens, _values, _states )
 result = val[0]
        result = val[1]
 result
end

def _reduce_36( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_37( tok, val, _tokens, _values, _states )
 result = val[0]
 result = val[1]
 result
end

def _reduce_38( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_39( tok, val, _tokens, _values, _states )
 result = val[0]
 result = val[1]
 result
end

def _reduce_40( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_41( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.routes.replace result.routes
        @field.addr = result.addr
 result
end

def _reduce_42( tok, val, _tokens, _values, _states )
 result = val[0]
 result = val
 result
end

def _reduce_43( tok, val, _tokens, _values, _states )
 result = val[0]
 result.push val[2]
 result
end

def _reduce_44( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_45( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_46( tok, val, _tokens, _values, _states )
 result = val[0]
 result = val
 result
end

def _reduce_47( tok, val, _tokens, _values, _states )
 result = val[0]
 result.push val[2]
 result
end

def _reduce_48( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_49( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_50( tok, val, _tokens, _values, _states )
 result = val[0]
        val[1].phrase = Bencode.decode( val[0] )
        result = val[1]
 result
end

def _reduce_51( tok, val, _tokens, _values, _states )
 result = val[0]
 result = AddrGroup.new( val[0], val[2] )
 result
end

def _reduce_52( tok, val, _tokens, _values, _states )
 result = val[0]
        val[2].route.replace val[1]
        result = val[2]
 result
end

def _reduce_53( tok, val, _tokens, _values, _states )
 result = val[0]
 result = val[1]
 result
end

def _reduce_54( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_55( tok, val, _tokens, _values, _states )
 result = val[0]
 result = [ val[1] ]
 result
end

def _reduce_56( tok, val, _tokens, _values, _states )
 result = val[0]
 result.push val[3]
 result
end

def _reduce_57( tok, val, _tokens, _values, _states )
 result = val[0]
 result = Mbox.new( val.join('') )
 result
end

def _reduce_58( tok, val, _tokens, _values, _states )
 result = val[0]
 result = Mbox.new( val[0] )
 result
end

def _reduce_59( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_60( tok, val, _tokens, _values, _states )
 result = val[0]
        result.concat val[1]
        result.concat val[2]
 result
end

def _reduce_61( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_62( tok, val, _tokens, _values, _states )
 result = val[0]
        result.concat val[1]
        result.concat val[2]
 result
end

def _reduce_63( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_64( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_65( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_66( tok, val, _tokens, _values, _states )
 result = val[0]
 result = '<' + val[1].addr + '>'
 result
end

def _reduce_67( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_68( tok, val, _tokens, _values, _states )
 result = val[0]
 result << ' ' << val[1]
 result
end

def _reduce_69( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_70( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_71( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_72( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_73( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_74( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.encrypter = val[0]
 result
end

def _reduce_75( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.encrypter = val[0]
        @field.keyword   = val[1]
 result
end

def _reduce_76( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.major = val[0].to_i
        @field.minor = val[2].to_i
 result
end

def _reduce_77( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.main = val[0]
        @field.sub  = val[2]
 result
end

def _reduce_78( tok, val, _tokens, _values, _states )
 result = val[0]
        @field.main = val[0]
        @field.sub  = ''
 result
end

def _reduce_79( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_80( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.params.store( val[2].downcase, val[4] )
 result
end

def _reduce_81( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_82( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_83( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.encoding = val[0]
 result
end

def _reduce_84( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.disposition = val[0]
 result
end

def _reduce_85( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_86( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_87( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_88( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.params.store( val[0].downcase, val[2] )
 result
end

def _reduce_89( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.params.store( val[0].downcase, val[2] )
 result
end

def _reduce_90( tok, val, _tokens, _values, _states )
 result = val[0]
 @field.params.store( val[0].downcase, val[2] )
 result
end

def _reduce_91( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_92( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_93( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_94( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_95( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_96( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end

def _reduce_97( tok, val, _tokens, _values, _states )
 result = val[0]

 result
end
##### racc generated code end #####
end
